window.Gleam.initSearch([{"doc":"chip","title":"chip","content":"# Chip - A subject registry for Gleam\n\n[![Package Version](https://img.shields.io/hexpm/v/chip)](https://hex.pm/packages/chip)\n[![Hex Docs](https://img.shields.io/badge/hex-docs-ffaff3)](https://hexdocs.pm/chip/)\n\nChip is capable of registering a set of\n[subjects](https://hexdocs.pm/gleam_erlang/gleam/erlang/process.html#Subject)\nas part of a group.\n\n## Example\n\nCategorize subjects in groups, then send messages to them:\n\n```gleam\nimport artifacts/game.{DrawCard, FireDice, PlayChip}\nimport chip\nimport gleam/list\n\npub type Group {\n  GroupA\n  GroupB\n}\n\npub fn main() {\n  let assert Ok(registry) = chip.start(chip.Unnamed)\n\n  let assert Ok(session_a) = game.start(DrawCard)\n  let assert Ok(session_b) = game.start(FireDice)\n  let assert Ok(session_c) = game.start(PlayChip)\n\n  chip.register(registry, GroupA, session_a)\n  chip.register(registry, GroupB, session_b)\n  chip.register(registry, GroupA, session_c)\n\n  chip.members(registry, GroupA, 50)\n  |&gt; list.each(fn(session) { game.next(session) })\n}\n```\n\nFor more check the [docs and guildelines](https://hexdocs.pm/chip/).\n\n## Development\n\nNew additions to the API will be considered with care. Features are\n[documented as Issues](https://github.com/chouzar/chip/issues?q=is%3Aopen+is%3Aissue+label%3Aenhancement)\non the project&#39;s repo, if you have questions or like to see a new feature please open an issue.\n\nRun tests:\n\n```sh\ngleam test\n```\n\nRun benchmarks:\n\n```sh\ngleam run --module benchmark\n```\n\n### Previous Art\n\nThis registry takes and combines some ideas from:\n\n* Elixir’s [registry](https://hexdocs.pm/elixir/Kernel.html) module.\n* Erlang&#39;s [pg](https://www.erlang.org/doc/apps/kernel/pg.html) module.\n* The [syn](https://github.com/ostinelli/syn) global registry library.\n\n### Other Gleam registry libraries\n\nOther registry libraries will provide different semantics and functionality:\n\n* [Singularity](https://hexdocs.pm/singularity/) is designed to register a fixed number of actors, where each one may have a different message type.\n\n## Installation\n\n```sh\ngleam add chip\n```\n","url":"index.html"},{"doc":"chip","title":"chip","content":"# Chip as your local PubSub system\n\nA PubSub will allow us to &quot;subscribe&quot; subjects to a &quot;topic&quot;, then later we\ncan &quot;publish&quot; an event to all subscribed subjects.\n\nThis pattern would be a generic way of re-defining chip as a pubsub system:\n\n```gleam\nimport chip\nimport gleam/erlang/process\nimport gleam/list\nimport gleam/otp/supervisor\n\npub type PubSub(message, channel) =\n  chip.Registry(message, channel)\n\npub fn start() {\n  chip.start(chip.Unnamed)\n}\n\npub fn childspec() {\n  supervisor.worker(fn(_param) { start() })\n  |&gt; supervisor.returning(fn(_param, pubsub) { pubsub })\n}\n\npub fn subscribe(\n  pubsub: PubSub(message, channel),\n  channel: channel,\n  subject: process.Subject(message),\n) -&gt; Nil {\n  chip.register(pubsub, channel, subject)\n}\n\npub fn publish(\n  pubsub: PubSub(message, channel),\n  channel: channel,\n  message: message,\n) -&gt; Nil {\n  chip.members(pubsub, channel, 50)\n  |&gt; list.each(fn(subscriber) { process.send(subscriber, message) })\n}\n```\n\nIt may be used to wire-up applications that require reacting to events. For example,\nlets assume we want to create a chat application:\n\n```gleam\nimport artifacts/pubsub\nimport chip\nimport gleam/erlang/process\nimport gleam/int\nimport gleam/list\nimport gleam/otp/task\n\npub type Channel {\n  General\n  Coffee\n  Pets\n}\n\npub type Event {\n  Event(id: Int, message: String)\n}\n\npub fn main() {\n  let assert Ok(pubsub) = pubsub.start()\n\n  // For this scenario, out of simplicity, all clients are the current process.\n  let client = process.new_subject()\n\n  // Client is interested in coffee and pets.\n  chip.register(pubsub, Coffee, client)\n  chip.register(pubsub, Pets, client)\n\n  // Lets assume this is the server process broadcasting a welcome message.\n  task.async(fn() {\n    chip.members(pubsub, General, 50)\n    |&gt; list.each(fn(client) {\n      Event(id: 1, message: &quot;Welcome to General! Follow rules and be nice.&quot;)\n      |&gt; process.send(client, _)\n    })\n    chip.members(pubsub, Coffee, 50)\n    |&gt; list.each(fn(client) {\n      Event(id: 2, message: &quot;Ice breaker! Favorite cup of coffee?&quot;)\n      |&gt; process.send(client, _)\n    })\n    chip.members(pubsub, Pets, 50)\n    |&gt; list.each(fn(client) {\n      Event(id: 3, message: &quot;Pets!&quot;)\n      |&gt; process.send(client, _)\n    })\n  })\n\n  // It is then each client&#39;s responsability to listen to incoming messages,\n  // our previous client is only subscribed to coffee and pets, so it only receives those messages.\n  let assert True =\n    listen_for_messages(client, [])\n    |&gt; list.all(fn(message) {\n      case message {\n        &quot;2: Ice breaker! Favorite cup of coffee?&quot; -&gt; True\n        &quot;3: Pets!&quot; -&gt; True\n        _other -&gt; False\n      }\n    })\n}\n\nfn listen_for_messages(client, messages) -&gt; List(String) {\n  // This function will listen until messages stop arriving for 100 milliseconds.\n\n  // A selector is useful to transform our Events into types a client expects,\n  // in this case the client can only receive String messages so we cast\n  // events into strings with the `to_string` function.\n  let selector =\n    process.new_selector()\n    |&gt; process.selecting(client, to_string)\n\n  case process.select(selector, 100) {\n    Ok(message) -&gt;\n      // A message was received, capture it and attempt to listen for another message.\n      message\n      |&gt; list.prepend(messages, _)\n      |&gt; listen_for_messages(client, _)\n\n    Error(Nil) -&gt;\n      // A message was not received, stop listening and return captured messages in order.\n      messages\n      |&gt; list.reverse()\n  }\n}\n\nfn to_string(event: Event) -&gt; String {\n  int.to_string(event.id) &lt;&gt; &quot;: &quot; &lt;&gt; event.message\n}\n```\n\nThere are many ways to structure a PubSub system in Gleam so this guide is just a starting point.\n\nWhile building your system with a PubSub you may start getting into clashes between modules (gleam\ndoesn&#39;t allow circular dependencies), if you find yourself in this situation try to:\n\n* Divide shared types, for example put `Event` in its own module.\n* Take advantage of generics so you don&#39;t have to be bound to specific types.\n* Take advantage of callbacks so you don&#39;t have to be bound to specific behaviour.\n\nIf in doubt try to keep your types and functions in one module and separate as your abstractions\nand domain knowledge mature.\n","url":"guides/as-local-pubsub.html"},{"doc":"chip","title":"chip","content":"# Chip as a session index\n\nWe may use chip to track individual subjects through an identifier. The\nidentifier may be an integer, string or even a union type if you only\nneed few records.\n\nThis pattern would be a way of re-defining chip as an indexed store:\n\n```gleam\nimport chip\nimport gleam/erlang/process\nimport gleam/otp/actor\nimport gleam/otp/supervisor\n\npub type Store(message) =\n  chip.Registry(message, Int)\n\npub type Id =\n  Int\n\npub fn start() -&gt; Result(Store(message), actor.StartError) {\n  chip.start(chip.Unnamed)\n}\n\npub fn childspec() {\n  supervisor.worker(fn(_index) { start() })\n}\n\npub fn index(store: Store(message), id: Id, subject: process.Subject(message)) {\n  chip.register(store, id, subject)\n}\n\npub fn get(store: Store(message), id: Id) {\n  case chip.members(store, id, 50) {\n    [] -&gt; Error(Nil)\n    [subject] -&gt; Ok(subject)\n    [_, ..] -&gt; panic as &quot;Shouldn&#39;t insert more than 1 subject.&quot;\n  }\n}\n```\n\nIt may be used to retrieve information from out of bound subjects. For example,\nin a web app we may add the registry as part of our app context:\n\n```gleam\nimport artifacts/game.{DrawCard}\nimport artifacts/store\n\npub fn main() {\n  let assert Ok(sessions) = store.start()\n\n  let assert Ok(session_1) = game.start(DrawCard)\n  let assert Ok(session_2) = game.start(DrawCard)\n  let assert Ok(session_3) = game.start(DrawCard)\n\n  store.index(sessions, 1, session_1)\n  store.index(sessions, 2, session_2)\n  store.index(sessions, 3, session_3)\n\n  router(sessions, &quot;/resource/&quot;, 2)\n}\n\nfn router(sessions, url, id) {\n  case url, id {\n    &quot;/resource/&quot;, id -&gt; render_resource(sessions, id)\n    _other, _id -&gt; render_error()\n  }\n}\n\nfn render_resource(sessions, id) {\n  case store.get(sessions, id) {\n    Ok(session) -&gt; game.current(session)\n    Error(Nil) -&gt; render_error()\n  }\n}\n\nfn render_error() {\n  &quot;Print Error&quot;\n}\n```\n\nOf course, if our sessions die we can no longer reference them through the index.\nFor process restart strategies check the [supervision guideline](as-part-of-supervisor.html).\n","url":"guides/as-subject-index.html"},{"doc":"chip","title":"chip","content":"# Chip as part of a Supervision tree\n\nIn the wider erlang ecosystem a supervision tree defines a strategy to startup and monitor \nerlang processes, in case one of the supervised process shutdowns or fails, this same \nstrategy will be restarted, giving the system self-healing capabilities.\n\nTo make chip and other subjects part of a supervision tree we need to define their child\nspecifications, these specifications define their behaviour and state when starting or\nre-starting.\n\nLets assume we need to have multiple &quot;sessions&quot; indexed on our system:\n\n```gleam\nimport chip\nimport gleam/erlang/process\nimport gleam/otp/supervisor\n\npub fn main() {\n  let self = process.new_subject()\n  let assert Ok(_supervisor) = supervisor(self)\n\n  // Once initialized, the supervisor function will send back a message\n  // with the child registry. From then we can use the registry to\n  // find subjects.\n  let assert Ok(registry) = process.receive(self, 500)\n  let assert [_, _] = chip.members(registry, GroupA, 50)\n  let assert [_, _] = chip.members(registry, GroupB, 50)\n  let assert [_] = chip.members(registry, GroupC, 50)\n}\n\n// ------ Supervision Tree ------ //\n\n// A context type will help carry around state between children in the supervisor.\ntype Context {\n  Context(caller: process.Subject(Registry), registry: Registry, group: Group)\n}\n\n// The tree is defined by calling a hierarchy of specifications\nfn supervisor(main: process.Subject(Registry)) {\n  supervisor.start_spec(\n    supervisor.Spec(\n      argument: main,\n      max_frequency: 5,\n      frequency_period: 1,\n      init: fn(children) {\n        children\n        // First spawn the registry.\n        |&gt; supervisor.add(registry_spec())\n        // Then spawn all sessions.\n        |&gt; supervisor.add(session_spec())\n        |&gt; supervisor.add(session_spec())\n        |&gt; supervisor.add(session_spec())\n        |&gt; supervisor.add(session_spec())\n        |&gt; supervisor.add(session_spec())\n        // Finally notify the main process we&#39;re ready.\n        |&gt; supervisor.add(ready())\n      },\n    ),\n  )\n}\n\n// ------ Registry ------ //\n\ntype Registry =\n  chip.Registry(Message, Group)\n\nfn registry_spec() {\n  // The registry childspec first starts the registry.\n  supervisor.worker(fn(_caller: process.Subject(Registry)) {\n    chip.start(chip.Named(&quot;sessions&quot;))\n  })\n  // After starting we transform the parameter from caller into a context for\n  // the sessions we want to register.\n  |&gt; supervisor.returning(fn(caller, registry) {\n    Context(caller, registry, GroupA)\n  })\n}\n\n// ------ Session ------- //\n\nfn session_spec() {\n  supervisor.worker(fn(context: Context) {\n    start_session(context.registry, context.group)\n  })\n  |&gt; supervisor.returning(fn(context: Context, _game_session) {\n    // Increments the id for the next session.\n    Context(..context, group: next_group(context.group))\n  })\n}\n\nfn start_session(\n  with registry: Registry,\n  group group: Group,\n) -&gt; supervisor.StartResult(Message) {\n  // Mock function to startup a new session.\n  let session = process.new_subject()\n  chip.register(registry, group, session)\n  Ok(session)\n}\n\n// ------ Helpers ------ //\n\ntype Message =\n  Nil\n\ntype Group {\n  GroupA\n  GroupB\n  GroupC\n}\n\nfn next_group(group) {\n  case group {\n    GroupA -&gt; GroupB\n    GroupB -&gt; GroupC\n    GroupC -&gt; GroupA\n  }\n}\n\nfn ready() {\n  // This childspec is a noop addition to the supervisor, on return it\n  // will send back the registry reference.\n  supervisor.worker(fn(_context: Context) { Ok(process.new_subject()) })\n  |&gt; supervisor.returning(fn(context: Context, _self) {\n    process.send(context.caller, context.registry)\n    Nil\n  })\n}\n```\n","url":"guides/as-part-of-supervisor.html"},{"doc":"chip","title":"Message","content":"pub opaque type Message(msg, group)\n Chip&#39;s internal message type.\n\nSynonyms:\nchip.Message\nchip Message","url":"chip.html#Message"},{"doc":"chip","title":"Named","content":"pub type Named {\n  Named(String)\n  Unnamed\n}\n An option passed to `chip.start` to make the registry available through a name.\nNamed(String)\n\n\nUnnamed\n\n\nSynonyms:\nchip.Named\nchip Named","url":"chip.html#Named"},{"doc":"chip","title":"Registry","content":"pub type Registry(msg, group) =\n  Subject(Message(msg, group))\n A shorter alias for the registry&#39;s subject.\n\nSynonyms:\nchip.Registry\nchip Registry","url":"chip.html#Registry"},{"doc":"chip","title":"from","content":"pub fn from(name: String) -&gt; Result(Subject(Message(a, b)), Nil)\n Retrieves a previously named registry.\n\n ## Example\n\n This function can be useful when there is no registry subject in scope.\n\n \n let _ = chip.start(chip.Named(&quot;sessions&quot;))\n let assert Ok(registry) = chip.from(&quot;sessions&quot;)\n \n\n Be mindful that using `from` means you lose type safety as the original `group` and\n subject `message` will be not inferred from it. To circunvent this it is possible\n to manually specify the type signature.\n\n \n let assert Ok(registry) = chip.from(&quot;sessions&quot;)\n\n // specify through a typed variable\n let registry: chip.Registry(Message, Groups)\n\n // specify through a helper function\n fn get_session(name: String) -&gt; chip.Registry(Message, Groups) {\n   case chip.from(&quot;sessions&quot;) {\n     Ok(registry) -&gt; registry\n     Error(Nil) -&gt; panic as &quot;session is not available&quot;\n   }\n }\n \nSynonyms:\nchip.from\nchip from","url":"chip.html#from"},{"doc":"chip","title":"members","content":"pub fn members(\n  registry: Subject(Message(a, b)),\n  group: b,\n  timeout: Int,\n) -&gt; List(Subject(a))\n Retrieves all subjects from a given group, order of retrieved\n subjects is not guaranteed.\n\n ## Example\n\n \n let assert Ok(registry) = chip.start(chip.Unnamed)\n\n chip.register(registry, GroupA, subject)\n chip.register(registry, GroupB, subject)\n chip.register(registry, GroupA, subject)\n\n let assert [_, _] = chip.members(registry, GroupA, 50)\n let assert [_] = chip.members(registry, GroupB, 50)\n \nSynonyms:\nchip.members\nchip members","url":"chip.html#members"},{"doc":"chip","title":"register","content":"pub fn register(\n  registry: Subject(Message(a, b)),\n  group: b,\n  subject: Subject(a),\n) -&gt; Nil\n Registers a subject under a group.\n\n ## Example\n\n \n let assert Ok(registry) = chip.start(chip.Unnamed)\n\n chip.register(registry, GroupA, subject)\n chip.register(registry, GroupB, subject)\n chip.register(registry, GroupC, subject)\n \n\n A subject may be registered under multiple groups but it may only be\n registered one time on each group.\nSynonyms:\nchip.register\nchip register","url":"chip.html#register"},{"doc":"chip","title":"start","content":"pub fn start(\n  named: Named,\n) -&gt; Result(Subject(Message(a, b)), StartError)\n Starts the registry.\n\n ## Example\n\n The registry may be started in an unnamed fashion.\n\n \n &gt; let assert Ok(registry) = chip.start(chip.Unnamed)\n \n\n As a convenience, it is also possible to start a named registry which can be\n retrieved later by using [from](#from).\n\n \n &gt; let _ = chip.start(chip.Named(&quot;sessions&quot;))\n \nSynonyms:\nchip.start\nchip start","url":"chip.html#start"},{"doc":"chip","title":"stop","content":"pub fn stop(registry: Subject(Message(a, b))) -&gt; Nil\n Stops the registry.\n\n ## Example\n\n \n let assert Ok(registry) = chip.start(chip.Unnamed)\n chip.stop(registry)\n \nSynonyms:\nchip.stop\nchip stop","url":"chip.html#stop"},{"doc":"chip","title":"chip","content":" Chip is a local [subject](https://hexdocs.pm/gleam_erlang/gleam/erlang/process.html#Subject)\n registry that can store Gleam process subjects as part of a group. Will also\n automatically delist dead processes.","url":"chip.html"}]);